;;; init.el --- Emacs initiation file                 -*- lexical-binding: t -*-

;; Author: Case Duckworth <acdw@acdw.net>
;; Created: Sometime during Covid-19, 2020
;; Keywords: configuration
;; URL: https://tildegit.org/acdw/emacs
;; Bankruptcy: 8

;;; License:

;; Everyone is permitted to do whatever they like with this software
;; without limitation.  This software comes without any warranty
;; whatsoever, but with two pieces of advice:
;; - Be kind to yourself.
;; - Make good choices.

;;; Code:
(let ((early-features `((early-init . ,(locate-user-emacs-file "early-init"))
                        acdw private +key)))
  (dolist (feature early-features)
    (require (or (car-safe feature) feature) (cdr-safe feature) :noerror)))

(setup (:require +casing)
  (+casing-mode +1))

(setup (:require +emacs)
  ;; +emacs.el contains super-basic defaults that are basically necessary for
  ;; good functioning.  In this block, I add extra things or more "experimental"
  ;; ones that might not belong in a separate file.
  (:also-load +lisp)
  (:option truncate-string-ellipsis "…")
  ;; Bindings
  (:global "C-x C-k" #'kill-current-buffer
           "C-x 4 n" #'clone-buffer
           "C-c v" #'visible-mode
           "C-M-;" #'+lisp-comment-or-uncomment-sexp
           "M-j" nil
           "C-x o" (lambda () (interactive) (switch-to-buffer nil))
           "C-x C-o" #'+open-paragraph
           "C-w" #'+kill-word-backward-or-region
           ;; "C-x C-1" #'delete-other-windows
           ;; "C-x 2" #'+split-window-below-then
           ;; "C-x C-2" #'+split-window-below-then
           ;; "C-x 3" #'+split-window-right-then
           ;; "C-x C-3" #'+split-window-right-then
           )
  ;; C-h deletes backward - see https://idiomdrottning.org/bad-emacs-defaults
  (global-set-key (kbd "C-h") 'delete-backward-char)
  (keyboard-translate ?\C-h ?\C-?)
  ;; Hooks
  ;; Advice
  ;; https://old.reddit.com/r/emacs/comments/rlli0u/whats_your_favorite_defadvice/hph14un/
  (define-advice keyboard-escape-quit (:around (fn &rest r))
    "Don't close splits on `keyboard-escape-quit'."
    (let ((buffer-quit-function #'ignore))
      (apply fn r))))

(setup (:require +init)
  (:bind "C-c s" #'+init-sort-then-save)
  (:hook #'+init-add-setup-to-imenu))

(setup (:require auth-source)
  (:option auth-sources (list (private/ "authinfo"))))

(setup (:require auto-insert)
  (setf (alist-get '("lisp/.*\\.el\\'" . "+Emacs lisp")
                   auto-insert-alist nil nil #'equal)
        `(nil ";;; " (file-name-nondirectory (buffer-file-name))
              (make-string (max 2 (- fill-column (current-column) 27)) 32)
              "-*- lexical-binding: t; -*-" '(setq lexical-binding t)
              ,(concat "\n\n;;; Commentary:"
                       "\n\n;; ") _
              ,(concat "\n\n;;; Code:"
                       "\n\n"
                       "\n\n(provide '")
              (file-name-base (buffer-file-name))
              ,(concat ")" "\n;;; ")
              (file-name-nondirectory (buffer-file-name)) " ends here\n"))
  (auto-insert-mode +1))

(setup (:require cus-edit)
  ;; I don't use Custom to actually /make/ any customizations, but it's handy to
  ;; (A) see what options are available and (B) persist some changes across
  ;; restarts, for example, `safe-local-variables'.
  (:require +cus-edit)
  (:option custom-file (private/ "custom.el")
           custom-magic-show nil
           custom-magic-show-button t
           custom-raised-buttons nil
           custom-unlispify-tag-names nil
           custom-variable-default-form 'lisp)
  (dolist (var '(safe-local-variable-values
                 warning-suppress-types))
    (add-to-list '+custom-variable-allowlist var))
  (+custom-load-ignoring-most-customizations)
  (advice-add #'custom-buffer-create-internal :after #'+cus-edit-expand-widgets)
  (:with-mode Custom-mode
    (:local-set imenu-generic-expression +cus-edit-imenu-generic-expression)))

(setup (:require goto-addr)
  (if (fboundp #'global-goto-address-mode)
      (global-goto-address-mode)
    (add-hook 'after-change-major-mode-hook #'goto-address-mode)))

(setup (:require pulse)
  (:also-load +pulse)
  (:option pulse-flag nil
           pulse-delay 0.5
           pulse-iterations 1)
  (dolist (command '(+ace-window-or-switch-buffer
                     pop-mark pop-globl-mark))
    (add-to-list '+pulse-location-commands command))
  (+ensure-after-init #'+pulse-location-mode))

(setup (:require reading)
  ;;(:hook-into view-mode)                ; XXX doesn't go back
  )

(setup (:require user-save)
  (add-hook 'user-save-hook #'+clean-empty-lines)
  (user-save-mode +1))

(setup +key
  (+ensure-after-init #'+key-global-mode))

(setup abbrev
  (:option abbrev-file-name (sync/ "abbrev.el")
           save-abbrevs 'silent)
  (:hook-into text-mode
              circe-chat-mode))

(setup autorevert
  (:option global-auto-revert-non-file-buffers t
           auto-revert-verbose nil)
  (global-auto-revert-mode +1))

(setup browse-url
  (:require +browse-url)
  (:option
   browse-url-browser-function #'eww-browse-url
   +browse-url-browser-function browse-url-browser-function
   browse-url-generic-program (seq-some #'executable-find
                                        '("firefox"
                                          "chromium"
                                          "chrome"))
   browse-url-generic-args (seq-some (lambda (e)
                                       (when (equal (executable-find (car e))
                                                    browse-url-generic-program)
                                         (cdr e)))
                                     '(("firefox" "--new-tab")))
   browse-url-secondary-browser-function (if (executable-find "firefox")
                                             #'browse-url-firefox
                                           #'browse-url-default-browser)
   browse-url-new-window-flag nil
   browse-url-firefox-arguments '("--new-tab")
   browse-url-firefox-new-window-is-tab t)
  (defvar +invidious-host
    ;; TODO: Add variables for other transformations and what-not.
    "invidious.snopyta.org"
    "Host for invidious instance.")
  ;; Set up external browsing URLs.
  (add-to-list '+custom-variable-allowlist
               '+browse-url-secondary-browser-regexps)
  (dolist (domain '("github.com" "twitch.tv"
                    "gitlab.com"
                    "google.com"
                    "imgur.com"
                    "pixelfed" "instagram.com" "bibliogram.art"
                    "reddit.com" "teddit.net"
                    "taskiq"
                    "twitter.com" "nitter.net" "t.co"
                    "streamable.com" "spotify.com"
                    "hetzner.cloud"))
    (add-to-list '+browse-url-secondary-browser-regexps
                 (replace-regexp-in-string "\\." "\\\\." domain)))
  ;; Set up URL handlers.
  (+browse-url-set-handlers
   (list
    (cons (rx                           ; images
           "." (or "jpeg" "jpg" "png") eos)
          (lambda (&rest args)
            (apply
             (cond ((executable-find "mpv") #'+browse-image-with-mpv)
                   (t #'eww-browse-url))
             args)))
    (cons (rx (or ;; videos
               "youtube.com" "youtu.be" "invidious" "yewtu.be"
               (seq "." (or "mp4" "gif" "mov" "MOV" "webm") eos)
               ;; music
               "soundcloud.com" "bandcamp.com"
               (seq "." (or "ogg" "mp3") eos)))
          (lambda (&rest args)
            (apply (if (executable-find "mpv")
                       #'+browse-url-with-mpv
                     browse-url-secondary-browser-function)
                   args)))
    (cons (+browse-url-secondary-browser-regexps-combine) ; non-text websites
          browse-url-secondary-browser-function)
    (cons "."                           ; everything else
          +browse-url-browser-function)))
  ;; Transform URLs before passing to `browse-url'
  (:option +browse-url-transformations `((,(rx (or "youtube.com"
                                                   "youtu.be"))
                                          . ,+invidious-host)
                                         ("twitter\\.com"
                                          . "nitter.net")
                                         ("instagram\\.com"
                                          . "bibilogram.art")
                                         (,(rx (or "reddit.com"
                                                   "old.reddit.com"))
                                          . "teddit.net")
                                         ("medium\\.com"
                                          . "scribe.rip")))
  (+browse-url-transform-url-global-mode +1))

(setup calendar
  (require '_location)
  (:option calendar-location-name _location-name
           calendar-latitude _location-latitude
           calendar-longitude _location-longitude
           diary-file (private/ "diary")))

(setup compile
  (:option compilation-always-kill t
           compilation-ask-about-save nil
           compilation-scroll-output t))

(setup dired
  (:also-load dired-x)
  (:also-straight dired-subtree
                  dired-collapse
                  dired-git-info
                  dired+)
  (:option dired-recursive-copies 'always
           dired-recursive-deletes 'always
           dired-create-destination-dirs 'always
           dired-do-revert-buffer t
           dired-hide-details-hide-symlink-targets nil
           dired-isearch-filenames 'dwim
           delete-by-moving-to-trash t
           dired-auto-revert-buffer t
           dired-listing-switches "-Al"
           ls-lisp-dirs-first t
           dired-ls-F-marks-symlinks t
           dired-clean-confirm-killing-deleted-buffers nil
           dired-no-confirm '(byte-compile
                              load chgrp chmod chown
                              copy move hardlink symlink
                              shell touch)
           dired-dwim-target t)
  (:local-set truncate-lines t)
  (:bind "<backspace>" #'dired-up-directory
         "TAB" #'dired-subtree-cycle
         "i" #'dired-subtree-toggle
         ")" #'dired-git-info-mode)
  (:hook #'dired-collapse-mode
         #'dired-hide-details-mode
         #'hl-line-mode)
  (:+key "C-x C-j" #'dired-jump)
  (with-eval-after-load 'dired
    (pcase system-system
      ('windows
       (:straight w32-browser))
      ('linux
       (:straight dired-open)
       (:option dired-listing-switches
                (concat dired-listing-switches " -F")))))
  (with-eval-after-load 'frowny
    (add-to-list 'frowny-inhibit-modes #'dired-mode)))

(setup eldoc
  (:hook-into elisp-mode
              lisp-interaction-mode))

(setup elisp-mode
  (:also-load +elisp)
  (:option eval-expression-print-length nil
           eval-expression-print-level nil)
  (:with-map (emacs-lisp-mode-map lisp-interaction-mode-map)
    (:bind "C-c C-c" #'eval-defun
           "C-c C-k" #'+elisp-eval-region-or-buffer
           "C-c C-z" #'ielm))
  (advice-add #'eval-region :around #'+eval-region@pulse))

(setup eshell
  (:also-load em-smart
              em-tramp)
  (:require +eshell)
  (+define-dir eshell/ (locate-user-emacs-file "eshell")
    "Where to place Eshell-specific files.")
  (:option eshell-aliases-file (eshell/ "aliases")
           ;; What are these for???
           eshell-rc-script (eshell/ "profile")
           eshell-login-script (eshell/ "login")
           eshell-destroy-buffer-when-process-dies t
           eshell-directory-name eshell/
           eshell-error-if-no-glob t
           eshell-hist-ignore-dups t
           eshell-kill-on-exit nil
           eshell-prefer-lisp-functions t
           eshell-prefer-lisp-variables t
           eshell-review-quick-commands nil
           eshell-save-history-on-exit t
           eshell-scroll-to-bottom-on-input 'all
           eshell-smart-space-goes-to-end t
           eshell-where-to-jump 'begin
           eshell-banner-message "")
  (with-eval-after-load 'mwim
    (setf (alist-get 'eshell-mode mwim-beginning-of-line-function)
          #'eshell-bol))
  (+eshell-eval-after-load
    ;; Local modes
    (dolist (mode '((hungry-delete-mode . -1)))
      (funcall (car mode) (cdr mode)))
    ;; Set local settings
    (dolist (setting (list (cons 'outline-regexp eshell-prompt-regexp)
                           (cons 'page-delimiter eshell-prompt-regexp)
                           (cons 'imenu-generic-expression
                                 (list "Prompt"
                                       (concat eshell-prompt-regexp
                                               "\\(.*\\)")
                                       1))))
      (set (make-local-variable (car setting)) (cdr setting)))
    ;; Bind keys
    (dolist (binding '(("C-d" . +eshell-quit-or-delete-char)))
      (define-key eshell-mode-map
                  (kbd (car binding)) (cdr binding)))
    ;; Environment variables
    (dolist (environment '(("PAGER" . "cat")))
      (setenv (car environment) (cdr environment)))))

(setup eww
  (:also-load +eww)
  (:option eww-search-prefix "https://duckduckgo.com/html?q="
           url-privacy-level '(email agent cookies lastloc))
  (add-hook 'eww-after-render-hook #'reading-mode)
  (:hook #'+eww-bookmark-setup
         #'+eww-track-readable-mode)
  (:bind "b" #'bookmark-set
         "B" #'bookmark-jump
         "M-n" nil
         "M-p" nil))

(setup flyspell
  (:hook-into org-mode))

(setup hideshow
  (:also-load +hideshow)
  (:with-mode hs-minor-mode
    (:hook-into prog-mode)
    (:bind "C-<tab>" #'+hs-cycle
           "C-S-<tab>" #'+hs-global-cycle
           ;; but y tho
           "C-S-<iso-lefttab>" #'+hs-global-cycle)))

(setup ibuffer
  (:also-load ibuf-ext)
  (:option ibuffer-expert t
           ibuffer-show-empty-filter-groups nil
           ibuffer-saved-filter-groups
           '(("default"
              ("Org" (mode . org-mode))
              ("emacs" (or (name . "^\\*scratch\\*$")
                           (name . "^\\*Messages\\*$")
                           (name . "^\\*Warnings\\*$")
                           (name . "^\\*straight-process\\*$")
                           (name . "^\\*Calendar\\*$")))
              ("customize" (mode . Custom-mode))
              ("emacs-config" (or (filename . ".emacs.d")
                                  (mode . +init-mode)))
              ("git" (or (name . "^\*magit")
                         (name . "^\magit")))
              ("help" (or (mode . help-mode)
                          (mode . Info-mode)
                          (mode . helpful-mode)))
              ("irc" (or (mode . erc-mode)
                         (mode . circe-server-mode)
                         (mode . circe-channel-mode)))
              ("shell" (or (mode . eshell-mode)
                           (mode . shell-mode)
                           (mode . vterm-mode)))
              ("web" (or (mode . elpher-mode)
                         (mode . eww-mode))))))
  (:hook (defun ibuffer@filter-to-default ()
           (ibuffer-auto-mode +1)
           (ibuffer-switch-to-saved-filter-groups "default"))))

(setup info
  (:also-load +Info)
  (:with-mode Info-mode                 ; -_-
    (:hook #'reading-mode)
    (:bind "c" #'+Info-copy-current-node-name
           "w" #'+Info-copy-current-node-name)))

(setup kmacro
  (:also-load +kmacro)
  (with-eval-after-load '+kmacro
    (+kmacro-recording-indicator-mode +1)
    (+kmacro-block-undo-mode +1)))

(setup magit
  ;; This setup is weird because of dependency issues
  (:straight (transient :host github :repo "magit/transient" :branch "master")
             (magit :host github :repo "magit/magit")
             (git-modes :host github :repo "magit/git-modes")))

(setup minibuffer
  (:require +minibuffer)
  (:with-map minibuffer-local-map
    (:bind "M-/" #'+minibuffer-complete-history)))

(setup mouse
  ;; Brand new for Emacs 28: see https://ruzkuku.com/texts/emacs-mouse.html
  ;; Actually, look at this as well: https://www.emacswiki.org/emacs/Mouse3
  (when (fboundp 'context-menu-mode)
    (:option context-menu-functions
             '(context-menu-ffap
               context-menu-region
               context-menu-undo
               context-menu-dictionary))
    (context-menu-mode +1))
  (dolist (click '(;; Fix scrolling in the margin
                   wheel-down double-wheel-down triple-wheel-down
                   wheel-up double-wheel-up triple-wheel-up))
    (global-set-key (vector 'right-margin click) 'mwheel-scroll)
    (global-set-key (vector 'left-margin click) 'mwheel-scroll)))

(setup org
  ;; Plain org with the `setup' form for sorting, but I install with straight.
  (:straight (org
              :type git :host nil
              :repo "https://git.savannah.gnu.org/git/emacs/org-mode.git"
              :local-repo "org"
              :depth full
              :pre-build (straight-recipes-org-elpa--build)
              :build (:not autoloads)
              :files (:defaults
                      "lisp/*.el"
                      ("etc/styles/" "etc/styles/*")))
             (org-contrib
              :type git :host nil
              :repo "https://git.sr.ht/~bzg/org-contrib"))
  ;; DO NOT load system-installed org !!!
  (setq load-path (cl-remove-if (lambda (path)
                                  (string-match-p "lisp/org\\'" path))
                                load-path))
  (:also-load +org
              _work)
  (:option org-adapt-indentation nil
           org-archive-mark-done t
           org-catch-invisible-edits 'show-and-error
           org-clock-clocked-in-display 'mode-line
           org-clock-frame-title-format (cons
                                         '(t org-mode-line-string)
                                         (cons " --- " frame-title-format))
           org-clock-string-limit 7     ; just the clock bit
           ;; org-clock-string-limit 25    ; gives enough information
           org-clock-persist t
           org-confirm-babel-evaluate nil
           org-cycle-separator-lines 0
           org-directory (sync/ "org/" t)
           org-ellipsis truncate-string-ellipsis
           org-fontify-done-headline t
           org-fontify-quote-and-verse-blocks t
           org-fontify-whole-heading-line t
           org-hide-emphasis-markers t
           org-html-coding-system 'utf-8-unix
           org-image-actual-width (list (* (window-font-width)
                                           (- fill-column 8)))
           org-imenu-depth 3
           org-list-demote-modify-bullet '(("-" . "+")
                                           ("+" . "-"))
           org-log-done 'time
           org-log-into-drawer t
           org-outline-path-complete-in-steps nil
           org-pretty-entities t
           org-pretty-entities-include-sub-superscripts nil
           org-refile-targets '((nil . (:maxlevel . 2))
                                (org-agenda-files . (:maxlevel . 1)))
           org-refile-use-outline-path 'file
           org-special-ctrl-a/e t
           org-special-ctrl-k t
           org-src-fontify-natively t
           org-src-tab-acts-natively t
           org-src-window-setup 'current-window
           org-startup-truncated nil
           org-startup-with-inline-images t
           org-tags-column (- (- fill-column (length org-ellipsis)))
           org-todo-keywords '((sequence "TODO(t)" "WAIT(w@/!)"
                                         "|" "DONE(d!)")
                               (sequence "|" "CANCELED(k!)")
                               (sequence "MEETING(m)")))
  (:bind "RET" #'+org-return-dwim
         "<S-return>" #'+org-table-copy-down
         "C-c C-l" #'+org-insert-link-dwim
         "C-c C-n" #'+org-next-heading-widen
         "C-c C-p" #'+org-previous-heading-widen
         "C-c C-o" #'+org-open-at-point-dwim
         "`" #'+org-insert-tilde
         "~" #'+org-insert-backtick)
  (:local-hook user-save-hook #'+org-before-save@prettify-buffer)
  (advice-add #'org-delete-backward-char :override #'+org-delete-backward-char)
  ;; (define-advice org-open-at-point (:around (fn &rest r) open-external)
  ;;   "Open links from org externally."
  ;;   (let ((browse-url-browser-function browse-url-secondary-browser-function))
  ;;     (apply fn r)))
  (with-eval-after-load 'org
    (setf (alist-get "\\.x?html?\\'" org-file-apps nil nil #'equal)
          #'+org-open-html)
    (org-clock-persistence-insinuate)
    (org-link-set-parameters "tel" :follow #'+org-tel-open)
    (setf (alist-get "\\.x?html?\\'" org-file-apps nil nil #'equal)
          #'+org-open-html))
  ;; Extra keywords
  (font-lock-add-keywords
   'org-mode
   '(;; Fancy list bullets
     ("^ *\\([-]\\) "
      (0 (compose-region (match-beginning 1) (match-end 1) "•")))
     ("^ *\\([+]\\) "
      (0 (compose-region (match-beginning 1) (match-end 1) "◦")))))
  (with-eval-after-load 'form-feed
    ;; Horizontal lines
    (font-lock-add-keywords
     'org-mode
     '(("^-----+" . form-feed--font-lock-face)))))

(setup org-agenda
  (:option org-agenda-skip-deadline-if-done t
           org-agenda-skip-scheduled-if-done t
           org-agenda-span 10
           org-agenda-include-diary nil ; I use the org-diary features
           org-agenda-todo-ignore-deadlines 'near
           org-agenda-todo-ignore-scheduled 'future
           org-agenda-include-deadlines t
           org-deadline-warning-days 0
           org-agenda-show-future-repeats 'next
           org-agenda-window-setup 'current-window)
  (dolist (var '(org-agenda-files
                 org-agenda-file-regexp
                 org-agenda-templates))
    (add-to-list '+custom-variable-allowlist var))
  (with-eval-after-load 'org
   (add-to-list 'org-agenda-files (sync/ "org/" t)))
  (:+leader "a" #'org-agenda "C-a" #'org-agenda)
  (:hook #'hl-line-mode)
  (add-hook 'org-agenda-after-show-hook 'org-narrow-to-subtree))

(setup org-capture
  (:require +org-capture)
  (:+leader "c" #'org-capture "C-c" #'org-capture)
  (+org-capture-templates-setf "t" "Todo")
  (+org-capture-templates-setf "tt"
      `("Today!" entry (file "todo.org")
        ,(concat "* TODO %^{Title}\n"
                 "DEADLINE: %t\n"
                 "\n%?")))
  (+org-capture-templates-setf "ts"
      `("Someday..." entry (file "todo.org")
        ,(concat "* TODO %^{Title}\n"
                 ":PROPERTIES:\n"
                 ":CREATED: [%<%F %T>]\n"
                 ":END:\n"
                 "\n%?")))
  (+org-capture-templates-setf "l"
      `("Link" entry (file "links.org")
        "* %(+org-insert-link-dwim) %^g\n\n"))
  (+org-capture-sort))

(setup org-contacts
  (:also-straight org-vcard)                 ; for importing Vcard files
  (:option org-contacts-matcher "contact")    ; Contacts are tagged "contact"
  )

(setup org-export
  (:also-load ox-md)
  (:option org-export-coding-system 'utf-8-unix
           org-export-headline-levels 8
           org-export-with-section-numbers nil
           org-export-with-smart-quotes t
           org-export-with-sub-superscripts t
           org-export-with-toc nil)
  (with-eval-after-load 'user-save
    (advice-add 'org-export-dispatch :before 'user-save-run-hooks)))

(setup password-cache
  (:option password-cache t
           password-cache-expiry (* 60 60)))

(setup prog
  (:local-set comment-auto-fill-only-comments t)
  (:hook #'prettify-symbols-mode
         #'display-fill-column-indicator-mode
         #'turn-on-auto-fill))

(setup scratch
  (:require +scratch)
  (:option initial-major-mode #'lisp-interaction-mode
           initial-scratch-message
           ";; ABANDON ALL HOPE YE WHO ENTER HERE\n\n")
  (add-hook 'kill-buffer-query-functions #'+scratch-immortal))

(setup shr
  (:option shr-width (- fill-column 5)  ; pad out for wide letters
           shr-use-fonts t))

(setup tab-bar
  (:require +tab-bar)
  (:option tab-bar-tab-name-function '+tab-bar-basename
           tab-bar-tab-name-truncated-max 20
           tab-bar-tab-name-ellipsis truncate-string-ellipsis
           tab-bar-show t)
  (tab-bar-mode +1)
  (if (version< emacs-version "28.0")
      (+tab-bar-misc-info-mode +1)
    (:option tab-bar-format '(+tab-bar-format-menu-bar
                              tab-bar-format-history
                              tab-bar-format-tabs
                              tab-bar-separator
                              tab-bar-format-add-tab
                              tab-bar-format-align-right
                              +tab-bar-misc-info
                              tab-bar-separator
                              ))))

(setup text
  (:hook #'turn-on-auto-fill))

(setup time
  ;; TODO: +display-time-mode with MOAR customizations
  (:option display-time-default-load-average nil
           display-time-mail-file :disable)
  (with-eval-after-load 'discord
    (with-eval-after-load '+tab-bar
      (:option display-time-string-forms
               '((propertize (format-time-string "%H:%M %a %e" now)
                             'face '+tab-bar-extra
                             'help-echo (discord-date-string))
                 load
                 (if mail
                     (concat " "
                             (propertize
                              display-time-mail-string
                              'display `(when (and display-time-use-mail-icon
                                                   (display-graphic-p))
                                          ,@display-time-mail-icon
                                          ,@(when (and
                                                   display-time-mail-face
                                                   (memq
                                                    (plist-get
                                                     (cdr display-time-mail-icon)
                                                     :type)
                                                    '(pbm xbm)))
                                              (let ((bg (face-attribute
                                                         display-time-mail-face
                                                         :background)))
                                                (when (stringp bg)
                                                  (list :background bg)))))
                              'face display-time-mail-face))
                   "")
                 " "))))

  (display-time-mode +1))

(setup tramp
  (el-patch-feature tramp)
  (with-eval-after-load 'tramp
    (el-patch-defun tramp-debug-buffer-command-completion-p (_symbol buffer)
      "A predicate for Tramp interactive commands.
They are completed by \"M-x TAB\" only in Tramp debug buffers."
      (with-current-buffer buffer
        (el-patch-wrap 2
          (save-restriction
           (widen)
           (string-equal (buffer-substring 1 10) ";; Emacs:")))))))

(setup (:straight 0x0)
  (:option 0x0-default-server 'ttm)
  (with-eval-after-load 'embark
    (define-key embark-region-map (kbd "U") #'0x0-dwim)))

(setup (:straight ace-window)
  (:require +ace-window)
  (:option aw-keys '(?a ?s ?d ?f ?g ?h ?j ?k ?l)
           aw-display-mode-overlay nil
           aw-scope 'frame)
  (:+key "M-o" #'+ace-window-or-switch-buffer)
  (:face aw-mode-line-face ((t (:foreground "red"))))
  (+ace-window-display-mode +1))

(setup (:straight (actually-selected-window
                   :host github
                   :repo "duckwork/actually-selected-window.el"))
  (actually-selected-window-mode +1))

(setup (:straight adaptive-wrap)
  (:with-mode adaptive-wrap-prefix-mode
    (:hook-into visual-column-mode)))

(setup (:straight anzu)
  (:option anzu-cons-mode-line-p nil)
  (:+key [remap query-replace] #'anzu-query-replace-regexp
         [remap query-replace-regexp] #'anzu-query-replace-regexp)
  (global-anzu-mode +1)
  (:bind-into isearch
    [remap isearch-query-replace] #'anzu-isearch-query-replace
    [remap isearch-query-replace-regexp] #'anzu-isearch-query-replace-regexp))

(setup (:straight-when (apheleia
                        :host github :repo "raxod502/apheleia")
         (eq system-type 'gnu/linux))
  (:require apheleia
            +apheleia)
  ;; (setf (alist-get 'prettier apheleia-formatters)
  ;;       '("npx" "prettier" "--stdin-filepath" filepath))
  (apheleia-global-mode +1))

(setup (:straight avy)
  (:require avy +avy)
  (:option avy-background t
           avy-lead-faces
           '(avy-lead-face
             avy-lead-face-1 avy-lead-face-1 avy-lead-face-1
             avy-lead-face-1 avy-lead-face-1 avy-lead-face-1))
  (:face avy-background-face
         ((t (:foreground "#888888"))))
  (:+key "M-j" #'avy-goto-char-timer)
  (:bind-into isearch
    "M-j" #'avy-isearch)
  (:when-loaded
    (setf (alist-get ?. avy-dispatch-alist) #'avy-action-embark)))

(setup (:straight (cape
                   :host github :repo "minad/cape"))
  (dolist (fn
           ;; All available cape capfs listed here. Add them to the front since
           ;; they're reversed with `add-to-list'.
           '(cape-file
             cape-dabbrev
             cape-keyword
             cape-abbrev
             cape-ispell
             ;;cape-dict))
             (add-to-list 'completion-at-point-functions fn :append)))))

(setup (:straight circe)
  (:require _circe
            +circe)
  (:also-load circe-chanop)
  (+ensure-after-init (lambda () (defalias 'irc '+irc "Start IRC.")))

  ;; Formatting options
  (:option
   ;; Messages between users
   circe-format-action (format (format "%%%ds* {nick} {body}"
                                       (- +circe-left-margin 2))
                               " ")
   circe-format-say (format "{nick:%1$d.%1$ds} | {body}"
                            (- +circe-left-margin 3))
   circe-format-self-action circe-format-action
   circe-format-self-say (replace-regexp-in-string "|" ">" circe-format-say)
   circe-format-notice (format "-{nick:%1$d.%1$ds}---{body}"
                               (- +circe-left-margin 4))
   circe-format-message (format (format "%%%ds@ *{nick}* {body}"
                                        (- +circe-left-margin 2))
                                " ")
   circe-format-message-action (replace-regexp-in-string "@" "*"
                                                         circe-format-message)
   circe-format-self-message (format (format "%%%ds> *{chattarget}* {body}"
                                             (- +circe-left-margin 2))
                                     " ")
   ;; Meta messages
   circe-format-server-channel-creation-time (+circe-format-meta
                                              (concat "Channel {channel}"
                                                      " created on {date}") t)
   circe-format-server-ctcp (+circe-format-meta
                             (concat "CTCP PING request to {target} from"
                                     " {userhost}: {body}"))
   circe-format-server-ctcp-ping-reply (+circe-format-meta
                                        (concat
                                         "CTCP PING reply to {target} from"
                                         " {userhost}: {body}"))
   circe-format-server-part (+circe-format-meta "PART {channel}: {reason}")
   circe-format-server-quit (+circe-format-meta "QUIT: {reason}")
   circe-format-server-quit-channel (+circe-format-meta
                                     "QUIT {channel}: {reason}")
   circe-format-server-join (+circe-format-meta "JOIN: {userinfo}")
   circe-format-server-join-in-channel (+circe-format-meta
                                        "JOIN {channel}: {userinfo}")
   circe-format-server-lurker-activity (+circe-format-meta
                                        "(JOINED {joindelta} ago)")
   circe-format-server-message (+circe-format-meta "{body}" t)
   circe-fromat-server-mode-change (+circe-format-meta
                                    (concat "MODE: {target} {change}"
                                            " by {setter} ({userhost})") t)
   circe-format-server-netmerge (+circe-format-meta
                                 (concat "NETMERGE: {split} at {date}"
                                         " (/WL to see who's still missing)") t)
   circe-format-server-netsplit (+circe-format-meta
                                 (concat "NETSPLIT: {split}"
                                         " (/WL to see who left)") t)
   circe-format-server-nick-change (+circe-format-meta
                                    "NICK WAS {old-nick} ({userhost})"
                                    "new-nick")
   circe-format-server-nick-regain (+circe-format-meta
                                    "NICK REGAINED: {old-nick} ({userhost})"
                                    "new-nick")
   circe-format-server-notice (+circe-format-meta "-SERVER NOTICE- {body}" t)
   circe-format-server-topic-time (+circe-format-meta
                                   "TOPIC SET BY {setter} on {topic-date}")
   circe-format-server-topic-time-for-channel (+circe-format-meta
                                               (concat
                                                "TOPIC ({channel}) SET BY"
                                                " {setter} on {topic-date}"))
   circe-format-server-whois-idle (+circe-format-meta "IDLE FOR {idle-duration}"
                                                      "whois-nick")
   circe-format-server-whois-idle-with-signon (+circe-format-meta
                                               (concat
                                                "IDLE FOR {idle-duration}"
                                                " (signon: {signon-date})")
                                               "whois-nick")
   circe-format-server-rejoin (+circe-format-meta
                               (concat "REJOIN: {userinfo} "
                                       "after {departuredelta}"))
   circe-format-server-topic (+circe-format-meta "TOPIC: {new-topic}")
   circe-prompt-string (format (format "%%%ds> "
                                       (- +circe-left-margin 2))
                               " "))

  (:option +circe-server-buffer-action (lambda (buf)
                                         (message "Connected to %s" buf))
           +circe-network-inhibit-autoconnect _circe-network-inhibit-autoconnect
           circe-network-options _circe-network-options
           circe-color-nicks-everywhere t
           circe-default-part-message "See You, Space Cowpokes . . ."
           circe-default-user user-real-login-name
           circe-reduce-lurker-spam t
           circe-server-auto-join-default-type :after-auth)
  (:bind "C-c C-p" #'circe-command-PART
         "C-c C-t" #'+circe-current-topic
         "C-l" #'lui-track-jump-to-indicator
         "C-<return>" #'+circe-chat@set-prompt)

  (advice-add #'circe-command-PART :after #'+circe-kill-buffer)
  (advice-add #'circe-command-QUIT :after #'+circe-quit@kill-buffer)
  (advice-add #'circe-command-GQUIT :after #'+circe-gquit@kill-buffer)

  (:with-mode circe-chat-mode
    (:local-set lui-input-function #'+lui-filter)
    (:hook #'enable-circe-color-nicks
           #'enable-circe-new-day-notifier
           #'+circe-chat@set-prompt
           ;; Filters
           #'+circe-F/C-mode
           ;; For some reason `+circe-shorten-url-mode' won't work right out of
           ;; the gate.
           (lambda () (run-at-time 0.25 nil #'+circe-shorten-url-mode)))
    (:bind "C-c C-s" #'circe-command-SLAP))

  (:with-mode lui-mode
    (:option lui-fill-column (+ fill-column +circe-left-margin)
             lui-fill-type nil
             lui-time-stamp-position 'right-margin
             lui-time-stamp-format "[ %H:%M"
             lui-track-behavior 'before-switch-to-buffer
             lui-track-indicator 'bar
             lui-fill-remove-face-from-newline nil
             lui-formatting-list `((,(+lui-make-formatting-list-rx "*")
                                    1 lui-strong-face)
                                   (,(+lui-make-formatting-list-rx "_")
                                    1 lui-emphasis-face)
                                   (,(+lui-make-formatting-list-rx "/")
                                    1 lui-emphasis-face))
             lui-autopaste-function
             (defun +0x0-upload-string (string)
               "Upload a string using 0x0."
               (with-temp-buffer
                 (insert string)
                 (0x0-upload-text (0x0--choose-server)))
                (current-kill 0)))
    (add-to-list '+pulse-location-commands #'lui-track-jump-to-indicator)
    (:hook #'visual-line-mode
           #'enable-lui-track
           #'visual-fill-column-mode
           (defun +disable-electric-pair-mode ()
             "Disable `electric-pair-mode' in the current buffer."
             (interactive)
             (electric-pair-local-mode -1))
           #'enable-lui-autopaste)
    (:local-set fringes-outside-margins t
                right-margin-width (length lui-time-stamp-format)
                scroll-margin 0
                scroll-step 1
                word-wrap t
                wrap-prefix (+string-repeat +circe-left-margin " ")
                line-number-mode nil
                column-number-mode nil
                file-percentage-mode nil
                visual-fill-column-extra-text-width
                (cons +circe-left-margin 0))
    (with-eval-after-load 'vertico-multiform
      (setf (alist-get 'lui-next-button-or-complete vertico-multiform-commands)
            '(flat))))

  (:with-mode tracking-mode
    (:option tracking-position 'before-modes)
    (add-to-list 'mode-line-misc-info
                 '(tracking-mode
                   tracking-mode-line-buffers)))

  (with-eval-after-load 'topsy
    (:option (append topsy-mode-functions)
             '(circe-channel-mode . +circe-current-topic)))

  (with-eval-after-load 'circe-color-nicks
    (add-hook 'modus-themes-after-load-theme-hook #'circe-nick-color-reset))
  (add-hook 'kill-emacs-hook #'+circe-quit-all@kill-emacs))

(setup (:straight clhs))

(setup (:straight consult)
  (:also-load +consult)
  ;; from Consult wiki
  (:option register-preview-delay 0
           register-preview-function #'consult-register-format
           xref-show-xrefs-function #'consult-xref
           xref-show-definitions-function #'consult-xref
           tab-always-indent 'complete
           completion-in-region-function #'consult-completion-in-region)
  (advice-add #'register-preview :override #'consult-register-window)
  (advice-add #'completing-read-multiple :override
              #'consult-completing-read-multiple)
  (dolist (binding '(;; C-c bindings (mode-specific-map)
                     ("C-c h" . consult-history)
                     ("C-c m" . consult-mode-command)
                     ("C-c b" . consult-bookmark)
                     ("C-c k" . consult-kmacro)
                     ;; C-x bindings (ctl-x-map)
                     ("C-x M-:" . consult-complex-command)
                     ("C-x b" . consult-buffer)
                     ("C-x 4 b" . consult-buffer-other-window)
                     ("C-x 5 b" . consult-buffer-other-frame)
                     ;; Custom M-# bindings for fast register access
                     ("M-#" . consult-register-load)
                     ("M-'" . consult-register-store)
                     ("C-M-#" . consult-register)
                     ;; Other custom bindings
                     ("M-y" . consult-yank-pop)
                     ("<f1> a" . consult-apropos)
                     ;; M-g bindings (goto-map)
                     ("M-g e" . consult-compile-error)
                     ("M-g f" . consult-flymake) ; or consult-flycheck
                     ("M-g g" . consult-goto-line)
                     ("M-g M-g" . consult-goto-line)
                     ("M-g o" . consult-outline) ; or consult-org-heading
                     ("M-g m" . consult-mark)
                     ("M-g k" . consult-global-mark)
                     ("M-g i" . consult-imenu)
                     ("M-g I" . consult-imenu-multi)
                     ;; M-s bindings (search-map)
                     ("M-s f" . consult-find)
                     ("M-s F" . consult-locate)
                     ("M-s g" . consult-grep)
                     ("M-s G" . consult-git-grep)
                     ("M-s r" . consult-ripgrep)
                     ("M-s l" . consult-line)
                     ("M-s L" . consult-line-multi)
                     ("M-s m" . consult-multi-occur)
                     ("M-s k" . consult-keep-lines)
                     ("M-s u" . consult-focus-lines)
                     ;; Isearch integration
                     ("M-s e" . consult-isearch-history)))
    (global-set-key (kbd (car binding)) (cdr binding)))
  (with-eval-after-load 'isearch-mode
    (dolist (binding '(("M-e" . consult-isearch-history)
                       ("M-s e" . consult-isearch-history)
                       ("M-s l" . consult-line)
                       ("M-s L" . consult-line-multi)))
      (define-key isearch-mode-map (car binding) (cdr binding))))
  (:+menu "b" #'consult-buffer
          "f" #'find-file)
  (with-eval-after-load 'org-mode
    (define-key org-mode-map "M-g o" 'consult-org-heading))
  (advice-add 'consult-yank-pop :after '+yank@indent)
  (with-eval-after-load 'consult
    (:option consult-narrow-key "<"
             consult-project-root-function '+consult-project-root)
    (consult-customize
     consult-theme
     :preview-key '(:debounce 0.2 any)
     consult-ripgrep consult-git-grep consult-grep
     consult-bookmark consult-recent-file consult-xref
     consult--source-file consult--source-project-file
     consult--source-bookmark
     :preview-key (kbd "M-."))
    (consult-history-to-modes ((minibuffer-local-map . nil)
                               (shell-mode-map . shell-mode-hook)
                               (term-mode-map . term-mode-hook)
                               (term-raw-map . term-mode-hook)
                               (comint-mode-map . comint-mode-hook)
                               (sly-mrepl-mode-map . sly-mrepl-hook)))
    (with-eval-after-load 'orderless
      (:option consult--regexp-compiler #'consult--orderless-regexp-compiler))))

(setup (:straight consult-dir)
  (:+key "C-x C-d" #'consult-dir)
  (:with-map vertico-map
    (:bind "C-x C-d" #'consult-dir
           "C-x C-j" #'consult-dir-jump-file)))

(setup (:straight crux)
  ;; yes it's silly I have an addon to this addon.
  (:also-load +crux)
  (:option crux-shell-func #'crux-eshell
           +crux-default-date-format "%F")
  (:global "C-o" #'crux-smart-open-line
           "C-x 4 t" #'crux-transpose-windows
           "M-w" #'+crux-kill-ring-save
           "C-k" #'crux-kill-and-join-forward
           "C-c d" #'+crux-insert-date-or-time)
  (:+leader "s"  #'crux-visit-shell-buffer)

  (el-patch-feature crux)
  (with-eval-after-load 'crux
    (el-patch-defun crux-reopen-as-root ()
      "Find file as root if necessary.

Meant to be used as `find-file-hook'.
See also `crux-reopen-as-root-mode'."
      (unless (or
               ;; This helps fix for `nov-mode', and possibly others.
               (el-patch-add (null buffer-file-name))
               (tramp-tramp-file-p buffer-file-name)
               (equal major-mode 'dired-mode)
               (not (file-exists-p (file-name-directory buffer-file-name)))
               (file-writable-p buffer-file-name)
               (crux-file-owned-by-user-p buffer-file-name))
        (crux-find-alternate-file-as-root buffer-file-name))))
  (crux-reopen-as-root-mode +1))

(setup (:straight dictionary)
  (:option dictionary-use-single-buffer t)
  (autoload 'dictionary-search "dictionary"
    "Ask for a word and search it in all dictionaries" t)
  (:hook #'reading-mode))

(setup (:straight (discord
                   :host github
                   :repo "davep/discord.el"
                   :fork (:repo "duckwork/discord.el"))))

(setup (:straight dumb-jump)
  (add-hook 'xref-backend-functions #'dumb-jump-xref-activate))

(setup (:straight electric-cursor)
  (:option electric-cursor-alist '((overwrite-mode . hbar)
                                   (god-local-mode . box)))
  (electric-cursor-mode +1))

(setup (:straight elfeed)
  (:require +elfeed)
  (+define-dir elfeed/ (sync/ "emacs/elfeed/" t))
  (:option
   elfeed-curl-program-name (executable-find "curl")
   elfeed-use-curl elfeed-curl-program-name
   elfeed-curl-extra-arguments '("--insecure")
   elfeed-enclosure-default-dir (cl-loop for dir in '("~/var/download/"
                                                      "~/Downloads/")
                                         if (file-exists-p dir)
                                         return dir)
   elfeed-search-filter "@1-month-ago +unread"
   elfeed-search-trailing-width 24
   elfeed-search-title-min-width 24
   elfeed-search-title-max-width 78
   elfeed-search-remain-on-entry t
   elfeed-show-unique-buffers t
   elfeed-db-directory (elfeed/ "db/" t))
  (:+key "C-x w" #'elfeed)
  (advice-add #'elfeed-search-fetch :after #'beginning-of-buffer)
  (:with-mode elfeed-search-mode
    (:bind "&" #'+elfeed-search-browse-generic)
    (:hook #'hl-line-mode)
    ;; https://old.reddit.com/r/emacs/comments/rlli0u/whats_your_favorite_defadvice/hphfh4e/
    (advice-add #'elfeed-search-update--force :after #'elfeed-db-save)
    (advice-add #'elfeed :before #'elfeed-db-load))
  (:with-mode elfeed-show-mode
    (:bind "SPC" #'+elfeed-scroll-up-command
           "S-SPC" #'+elfeed-scroll-down-command
           "&" #'+elfeed-show-browse-generic
           "RET" #'shr-browse-url)
    (:hook #'reading-mode)
    (+elfeed-update-async-mode +1)))

(setup (:straight elfeed-org)
  (:also-load +org-capture)
  (:option rmh-elfeed-org-files (list (elfeed/ "elfeed.org" t)))
  (elfeed-org)
  (+org-capture-templates-setf "f"
      `("Feed" entry
        (file+olp ,(car rmh-elfeed-org-files) "Feeds")
        "* %? %^g")))

(setup (:straight elpher))

(setup (:straight embark)
  (:require embark
            +embark)
  (:option prefix-help-command 'embark-prefix-help-command
           embark-keymap-prompter-key ";")
  (:+key "C-." #'embark-act
         "M-." #'embark-dwim
         "<f1> B" #'embark-bindings)
  (:with-map minibuffer-local-map
    (:bind "C-." #'embark-act
           "M-." #'embark-dwim))
  (:with-map embark-file-map
    (:bind "l" #'vlf)))

(setup (:straight embark-consult)
  (:load-after consult embark)
  (add-hook 'embark-collect-mode-hook #'consult-preview-at-point-mode))

(setup (:straight epithet)
  (dolist (hook '(Info-selection-hook
                  eww-after-render-hook
                  help-mode-hook
                  occur-mode-hook))
    (add-hook hook #'epithet-rename-buffer)))

(setup (:straight eros)
  (:hook-into emacs-lisp-mode
              lisp-interaction-mode))

(setup (:straight eshell-bookmark)
  (add-hook 'eshell-mode-hook #'eshell-bookmark-setup))

(setup (:straight eshell-syntax-highlighting)
  (:hook-into eshell-mode))

(setup (:straight eshell-vterm)
  (:load-after eshell)
  (defalias 'eshell/v 'eshell-exec-visual)
  (eshell-vterm-mode +1))

(setup (:straight-when exec-path-from-shell
         (eq system-type 'gnu/linux))
  (require 'exec-path-from-shell)
  (dolist (var '("SSH_AUTH_SOCK"
                 "SSH_AGENT_PID"
                 "GPG_AGENT_INFO"
                 "LANG"
                 "LC_CTYPE"
                 "XDG_CONFIG_HOME"
                 "XDG_CONFIG_DIRS"
                 "XDG_DATA_HOME"
                 "XDG_DATA_DIRS"
                 "XDG_CACHE_HOME"))
    (add-to-list 'exec-path-from-shell-variables var))
  (exec-path-from-shell-initialize))

(setup (:straight expand-region)
  (:also-load +expand-region)
  (:option expand-region-fast-keys-enabled nil)
  (:+key "C-=" #'er/expand-region
         "C--" #'+er/contract-or-negative-argument))

(setup (:straight (fill-sentences-correctly
                   :host github
                   :repo "duckwork/fill-sentences-correctly.el"))
  (fill-sentences-correctly-mode +1))

(setup (:straight (filldent
                   :host github
                   :repo "duckwork/filldent.el"))
  (:+key "M-q" #'filldent-dwim))

(setup (:straight flyspell-correct)
  (:load-after flyspell)
  (:also-load +flyspell-correct)
  (:option flyspell-correct--cr-key ";")
  (:bind-into flyspell
    "C-;" #'flyspell-correct-wrapper
    "<f7>" #'+flyspell-correct-buffer)
  (with-eval-after-load 'vertico-multiform
    (setf (alist-get 'flyspell-correct-wrapper vertico-multiform-commands) nil)))

(setup (:straight-when (forge
                        :host github :repo "magit/forge")
         (eq system-type 'gnu/linux))
  (require 'forge)
  (add-to-list 'forge-alist
               '("tildegit.org" "tildegit.org/api/v1" "tildegit.org"
                 forge-gitea-repository)))

(setup (:straight form-feed)
  (global-form-feed-mode +1))

(setup (:straight (frowny
                   :host github
                   :repo "duckwork/frowny.el"))
  (global-frowny-mode +1))

(setup (:straight gcmh)
  (:option gcmh-idle-delay 'auto)
  (gcmh-mode +1))

(setup (:straight-when geiser
         (progn
           (defvar +schemes
             (let (schemes)
               (dolist (scheme '(("scheme" . geiser-chez) ; chez
                                 ("petite" . geiser-chez) ; petite
                                 ("csi" . geiser-chicken) ; chicken
                                 ("gsi" . geiser-gambit)
                                 ("gosh" . geiser-gauche)
                                 ("guile" . geiser-guile)
                                 ("kawa" . geiser-kawa)
                                 ("mit-scheme" . geiser-mit)
                                 ("racket" . geiser-racket)
                                 ("stklos" . geiser-stklos)))
                 (when-let (binary (executable-find (car scheme)))
                   (push binary schemes)
                   ;; and install the proper helper package
                   (straight-use-package (cdr scheme))))
               (nreverse schemes)))
           +schemes))
  (:file-match (rx ".rkt" eos)
               (rx ".scm" eos)))

;; (setup (:straight god-mode)
;;   (setq god-mode-enable-function-key-translation nil)
;;   (:require god-mode
;;             +god-mode)
;;   (:+key "C-M-g" #'god-mode-all)
;;   (:with-mode god-local-mode
;;     (:bind "i" #'+god-mode-insert
;;            "a" nil)))

(setup (:straight helpful)
  (run-with-idle-timer 0.5 nil
                       'require 'helpful)
  (:+key "<f1> f" #'helpful-callable
         "<f1> v" #'helpful-variable
         "<f1> k" #'helpful-key
         "<f1> ." #'helpful-at-point))

(setup (:straight (hippie-completing-read
                   :host github
                   :repo "duckwork/hippie-completing-read"))
  (:+key "M-/" #'hippie-completing-read))

(setup (:straight hungry-delete)
  (:option hungry-delete-chars-to-skip " \t"
           hungry-delete-join-reluctantly nil)
  (+with-ensure-after-init
    (add-to-list 'hungry-delete-except-modes 'eshell-mode))
  (:bind-into paredit
    ;; I define these functions here because they really require both packages
    ;; to make any sense.  So, would I put them in `+hungry-delete' or
    ;; `+paredit' ?  There's no satisfactory answer.
    [remap paredit-backward-delete]
    (defun acdw/paredit-hungry-delete-backward (arg)
      (interactive "P")
      (if (looking-back "[ \t]" 1)
          (hungry-delete-backward (or arg 1))
        (paredit-backward-delete arg)))
    [remap paredit-forward-delete]
    (defun acdw/paredit-hungry-delete-forward (arg)
      (interactive "P")
      (if (looking-at "[ \t]")
          (hungry-delete-forward (or arg 1))
        (paredit-forward-delete arg))))
  (global-hungry-delete-mode +1))

(setup (:straight isearch-mb)
  ;; This complicatedness is an attempt to make it easier to add and
  ;; subtract `isearch-mb' bindings using the suggestions in the
  ;; project's README.
  (with-eval-after-load 'isearch-mb
    (dolist (spec '((isearch-mb--with-buffer
                     ("M-e" .  consult-isearch)
                     ("C-o" . loccur-isearch))
                    (isearch-mb--after-exit
                     ("M-%" . anzu-isearch-query-replace)
                     ("M-s l" . consult-line))))
      (let ((isearch-mb-list (car spec))
            (isearch-mb-binds (cdr spec)))
        (dolist (cell isearch-mb-binds)
          (let ((key (car cell))
                (command (cdr cell)))
            (when (fboundp command)
              (add-to-list isearch-mb-list command)
              (define-key isearch-mb-minibuffer-map (kbd key) command)))))))
  (isearch-mb-mode +1))

(setup (:straight-when keychain-environment
         (executable-find "keychain"))
  (keychain-refresh-environment))

(setup (:straight lacarte)
  (:+key "<f10>" #'lacarte-execute-menu-command)
  (with-eval-after-load 'vertico-multiform
    (setf (alist-get 'lacarte-execute-menu-command vertico-multiform-commands)
          '(buffer grid (vertico-sort-function . vertico-sort-length-alpha)))))

(setup (:straight (lin :host gitlab :repo "protesilaos/lin"))
  (require 'lin)
  (dolist (hook lin-foreign-hooks)
    (add-hook hook #'hl-line-mode)
    (add-hook hook #'lin-mode)))

(setup (:straight link-hint)
  (:require +link-hint)
  (+link-hint-setup-open-secondary)
  (:option link-hint-avy-style 'at-full)
  (:+key "M-l" +link-hint-map)
  (:with-map +link-hint-map
    (:bind "M-l" #'+link-hint-open-link "l" #'+link-hint-open-link
           "M-m" #'link-hint-open-multiple-links
           "m" #'link-hint-open-multiple-links
           "M-c" #'link-hint-copy-link "c" #'link-hint-copy-link)))

(setup (:straight macrostep)
  (:require macrostep)
  (:with-mode emacs-lisp-mode
    (:bind "C-c e" #'macrostep-expand)))

(setup (:straight marginalia)
  (marginalia-mode +1))

(setup (:straight markdown-mode)
  (:option markdown-hide-markup t)
  (add-to-list 'auto-mode-alist (cons (rx (or ".md" ".markdown" ".mdown")
                                          eos)
                                      'markdown-mode))
  (with-eval-after-load 'apheleia
    (when-let ((mdfmt-exe (executable-find "markdownfmt")))
      (setf (alist-get 'markdownfmt apheleia-formatters) mdfmt-exe)
      (setf (alist-get 'markdown-mode apheleia-mode-alist) 'markdownfmt)
      (setf (alist-get 'gfm-mode apheleia-mode-alist) 'markdownfmt))))

(setup (:straight mastodon)
  (:option mastodon-instance-url "https://tiny.tilde.website"
           mastodon-client--token-file (.etc "mastodon.plstore")
           mastodon-auth-source-file (car auth-sources)
           mastodon-tl--enable-proportional-fonts t
           mastodon-tl--show-avatars t)
  (:hook #'mastodon-async-mode
         #'reading-mode
         #'hl-line-mode
         #'lin-mode))

(setup (:straight minions)
  (minions-mode +1))

(setup (:straight (mode-line-bell
                   :host github :repo "purcell/mode-line-bell"
                   :fork (:host github :repo "duckwork/mode-line-bell")))
  (:option mode-line-bell-flash-time 0.1)
  (mode-line-bell-mode +1))

(setup (:straight (modus-themes
                   :host gitlab
                   :repo "protesilaos/modus-themes"))
  (require 'modus-themes (.etc "straight/build/modus-themes/modus-themes"))
  (:option modus-themes-mixed-fonts t
           modus-themes-bold-constructs t
           modus-themes-italic-constructs t
           modus-themes-headings '((t . (background))))
  (+sunrise-sunset 'modus-themes-load-operandi 'modus-themes-load-vivendi))

(setup (:straight mwim)
  (:require +mwim)
  (:option +mwim-passthrough-modes '(comint-mode
                                     eshell-mode
                                     vterm-mode))
  (:+key "C-a" #'+mwim-beginning-maybe
         "C-e" #'+mwim-end-maybe))

(setup (:straight orderless)
  (:also-load +orderless)
  (:option completion-styles '(substring orderless basic)
           completion-category-defaults nil
           completion-category-overrides
           '((file (styles partial-completion))
             (command (styles +orderless-with-initialism))
             (variable (styles +orderless-with-initialism))
             (symbol (styles +orderless-with-initialism)))
           orderless-component-separator #'orderless-escapable-split-on-space
           orderless-style-dispatchers '(+orderless-dispatch)))

(setup (:straight org-appear)
  (:option org-appear-autoemphasis t
           org-appear-autoentities t
           org-appear-autokeywords t
           org-appear-autolinks nil
           org-appear-autosubmarkers t
           org-appear-delay 0)
  (:hook-into org-mode))

(setup (:straight org-sticky-header)
  ;; (:hook-into org-mode)
  )

(setup (:straight org-visibility)
  ;; I think this is causing hangups on Windows .. ?
  (:option org-visibility-state-file (.etc "org-visibility")
           org-visibility-include-regexps '("\\.org\\'"))
  ;; (org-visibility-enable-hooks)
  )

(setup (:straight orglink)
  (global-orglink-mode +1))

(setup (:straight package-lint
                  package-lint-flymake)
  (add-hook 'emacs-mode-hook #'package-lint-flymake-setup)
  ;; Remove it from init.el files
  (add-hook '+init-mode-hook #'flymake-mode-off))

(setup (:straight paredit)
  (:bind "DEL" #'paredit-backward-delete
         "C-<backspace>" #'paredit-backward-kill-word
         "C-w" (lambda (arg) (interactive "P")
                 (+kill-word-backward-or-region arg
                                                #'paredit-backward-kill-word))
         "M-s" nil)
  (dolist (hook '(emacs-lisp-mode-hook
                  eval-expression-minibuffer-setup-hook
                  ielm-mode-hook
                  lisp-interaction-mode-hook
                  lisp-mode-hook
                  scheme-mode-hook))
    (add-hook hook #'enable-paredit-mode))
  (:also-load eldoc)
  (eldoc-add-command #'paredit-backward-delete #'paredit-close-round))

(setup (:straight paren-face)
  (:hook-into emacs-lisp-mode
              ielm-mode sly-repl-mode
              lisp-mode
              lisp-interaction-mode
              scheme-mode))

(setup (:straight-when pdf-tools
         ;; Ensure we can build `pdf-tools'
         (or (executable-find "gcc")
             (executable-find "g++")))
  (pdf-tools-install t))

(setup (:straight (shell-command+
                   :host nil
                   :repo "https://git.sr.ht/~pkal/shell-command-plus"))
  (:option shell-command-prompt "$ ")
  (:bind-into dired
    "M-!" 'shell-command+)
  (:+key "M-!" #'shell-command+))

(setup (:straight sicp))

(setup (:straight (simple-modeline
                   :host github :repo "gexplorer/simple-modeline"
                   :fork (:host github :repo "duckwork/simple-modeline")))
  (:require +modeline)
  (:option +modeline-minions-icon ";"
           simple-modeline-segments
           `(( ; left
              +modeline-ace-window-display
              +modeline-modified
              +modeline-buffer-name
              (lambda () (+modeline-vc ": "))
              ,(+modeline-concat
                '(+modeline-minions
                  +modeline-major-mode))
              +modeline-anzu
              )
             ( ; right
              (lambda ()
                (unless +tab-bar-misc-info-mode
                  (+modeline-concat
                   '(+modeline-track
                     simple-modeline-segment-misc-info))))

              simple-modeline-segment-process
              +modeline-text-scale
              ,(+modeline-concat
                '(+modeline-god-mode
                  +modeline-reading-mode
                  +modeline-narrowed)
                ",")
              +modeline-position)))
  (simple-modeline-mode +1))

(setup (:straight-when sly
         (defvar +lisp-bin (executable-find "sbcl")))
  (:also-load sly-autoloads
              +sly)
  (:option inferior-lisp-program +lisp-bin
           sly-kill-without-query-p t)
  (:with-feature sly-mrepl
    (dolist (key '("RET" "<return>"))
      (:bind key #'sly-mrepl-return-at-end))
    (:bind "C-c C-c" #'sly-mrepl-return)))

(setup (:straight smartscan)
  (:with-map smartscan-map
    (:bind "M-'" nil))
  (:hook-into prog-mode))

(setup (:straight (sophomore
                   :host github
                   :repo "duckwork/sophomore.el"))
  (sophomore-enable #'narrow-to-region)
  (sophomore-disable                    ; These are mostly annoying commands
   #'view-hello-file
   #'describe-gnu-project)
  (sophomore-mode +1))

(setup (:straight ssh-config-mode)
  (:file-match (rx "/.ssh/config" eos)
               (rx "/ssh" (? "d") "_config" eos))
  (:with-mode ssh-known-hosts-mode
    (:file-match (rx "/knownhosts" eos)))
  (:with-mode ssh-authorized-keys-mode
    (:file-match (rx "/authorized_keys" (? "2") eos))))

(setup (:straight super-save)
  (:option auto-save-default nil
           super-save-auto-save-when-idle t
           super-save-idle-duration 60
           super-save-exclude '(".gpg")
           super-save-remote-files nil)
  (auto-save-visited-mode -1)
  (super-save-mode +1))

(setup (:straight-when systemd
         (executable-find "systemd"))
  (:option systemd-man-function 'woman))

(setup (:straight (titlecase
                   :host github
                   :repo "duckwork/titlecase.el"
                   :files ("*")))
  (:require titlecase)
  (:with-map +casing-map
    (:bind "t" #'titlecase-dwim
           "M-t" #'titlecase-dwim)))

(setup (:straight topsy)
  (:hook-into ;;prog-mode
   circe-chat-mode)
  (:when-loaded
    (:option
     topsy-header-line-format
     '(:eval
       (list
        (propertize " "
                    'display
                    `((space
                       :align-to
                       ,(unless (bound-and-true-p visual-fill-column-mode)
                          0))))
        (funcall topsy-fn))))))

(setup (:straight transpose-frame)
  (defvar +transpose-frame-map
    (let ((map (make-sparse-keymap)))
      (dolist (bind '(("t" . transpose-frame)
                      ("v" . flip-frame)
                      ("h" . flop-frame)
                      ("r" . rotate-frame-clockwise)
                      ("R" . rotate-frame-anticlockwise)))
        (define-key map (car bind) (cdr bind)))
      map)
    "Map for transposing frames.")
  (define-key +key-mode-map (kbd "C-x 5 t") +transpose-frame-map))

(setup (:straight trashed)
  (:option trashed-action-confirmer #'y-or-n-p))

(setup (:straight undo-fu)
  (:+key "C-/" #'undo-fu-only-undo
         "C-?" #'undo-fu-only-redo))

(setup (:straight undo-fu-session)
  (:option undo-fu-session-incompatible-files '("/COMMIT_EDITMSG\\'"
                                                "/git-rebase-todo\\'")
           undo-fu-session-directory (.etc "undo/" t)
           undo-fu-session-compression (executable-find "gzip"))
  (global-undo-fu-session-mode +1))

(setup (:straight (vertico
                   :host github
                   :repo "minad/vertico"
                   :files ("*" "extensions/*"
                           (:exclude ".git"))))
  (:also-load +vertico)
  (:option resize-mini-windows 'grow-only
           vertico-count-format nil
           vertico-cycle t)
  (advice-add #'vertico-next :around #'+vertico-ding-wrap)
  (when (boundp 'native-comp-deferred-compilation-deny-list)
    (add-to-list 'native-comp-deferred-compilation-deny-list "vertico"))
  (vertico-mode +1)
  ;; Extensions
  (:also-load vertico-directory
              vertico-mouse
              vertico-unobtrusive
              vertico-multiform
              vertico-quick)
  (vertico-mouse-mode +1)
  (vertico-multiform-mode +1)
  (:option vertico-multiform-commands '((execute-extended-command flat)
                                        (helpful-function)
                                        (helpful-macro)
                                        (helpful-callable)
                                        (helpful-variable)
                                        (describe-symbol)
                                        (completion-at-point flat)
                                        (indent-for-tab-command flat)
                                        (consult-buffer flat)
                                        (insert-char))
           ;; This is applied /after/ the above, so default is at the end of
           ;; this alist.
           vertico-multiform-categories '((file buffer grid)
                                          (t flat)))
  (dolist (buf-cmd '(consult-find
                     consult-yank-pop
                     consult-locate
                     consult-grep
                     consult-git-grep
                     consult-ripgrep
                     consult-line
                     consult-line-multi
                     consult-multi-occur
                     consult-keep-lines
                     consult-focus-lines
                     consult-imenu
                     consult-outline))
    (setf (alist-get buf-cmd vertico-multiform-commands) '(buffer)))
  (:with-map vertico-map
    (:bind "RET" #'vertico-directory-enter
           "DEL" #'vertico-directory-delete-char
           "M-DEL" #'vertico-directory-delete-word
           "TAB" #'+vertico-widen-or-complete
           "M-j" #'vertico-quick-insert))
  (add-hook 'rfn-eshadow-update-overlay-hook #'vertico-directory-tidy))

(setup (:straight visual-fill-column)
  (:option visual-fill-column-center-text t
           (append reading-modes) '(visual-fill-column-mode . +1))
  (:hook #'visual-line-mode)
  (:hook-into org-mode)
  (advice-add #'text-scale-adjust :after #'visual-fill-column-adjust))

(setup (:straight vlf)
  (:require vlf-setup))

(setup (:straight-when vterm
         (and module-file-suffix
              (executable-find "cmake")))
  (:also-load +vterm)
  (:option vterm-always-compile-module t
           vterm-buffer-name-string "vterm: %s"
           vterm-max-scrollback 100000  ; max allowed by vterm-module.h
           )
  (advice-add 'counsel-yank-pop-action :around
              #'+vterm-counsel-yank-pop-action))

(setup (:straight web-mode)
  (setf (alist-get (rx "." (or "htm" "html" "phtml" "tpl.php"
                               "asp" "gsp" "jsp" "ascx" "aspx"
                               "erb" "mustache" "djhtml")
                       eos)
                   auto-mode-alist)
        'web-mode)
  (with-eval-after-load 'apheleia
    (setf (alist-get 'web-mode apheleia-mode-alist)
          'prettier)))

(setup (:straight whitespace-cleanup-mode)
  (:option whitespace-cleanup-mode-preserve-point t
           whitespace-cleanup-mode-only-if-initially-clean nil)
  (global-whitespace-cleanup-mode +1))

(setup (:straight wrap-region)
  (:require wrap-region)
  (wrap-region-add-wrappers
   '(("*" "*" nil org-mode)
     ("~" "~" nil org-mode)
     ("/" "/" nil org-mode)
     ("=" "=" nil org-mode)
     ("+" "+" nil org-mode)
     ("_" "_" nil org-mode)
     ("$" "$" nil (org-mode latex-mode))))
  (:hook-into org-mode
              latex-mode))

(setup (:straight yasnippet)
  (:option yas-snippet-dirs (list
                             (expand-file-name "snippets" user-emacs-directory)
                             (sync/ "emacs/snippets" t)))
  (yas-global-mode +1))

(setup (:straight (ytdious
                   :host github :repo "spiderbit/ytdious"
                   :fork (:host github :repo "duckwork/ytdious")))
  (:also-load +ytdious)
  (:option ytdious-invidious-api-url (if +invidious-host
                                         (concat "https://" +invidious-host)
                                       "https://invidious.snopyta.org"))
  (:bind "y" #'+ytdious-watch))

(setup (:straight zoom-frm))

(setup (:straight zzz-to-char)
  (:require +zzz-to-char)
  (:option zzz-to-char-reach (+bytes 1 :kib))
  (:global "M-z" #'+zzz-to-char))
